﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using HeuristicLab.Algorithms.Bandits;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.GrammaticalOptimization {
  // SARSA (fig. 6.9 in Sutton & Barto)
  public class TemporalDifferenceTreeSearchSampler {
    private class TreeNode {
      public string ident;
      public int randomTries;
      public double q;
      public int tries;
      public TreeNode[] children;
      public bool done = false;

      public TreeNode(string id) {
        this.ident = id;
      }

      public override string ToString() {
        return string.Format("Node({0} tries: {1}, done: {2})", ident, tries, done);
      }
    }


    public event Action<string, double> FoundNewBestSolution;
    public event Action<string, double> SolutionEvaluated;

    private readonly int maxLen;
    private readonly IProblem problem;
    private readonly Random random;
    private readonly int randomTries;
    private readonly IBanditPolicy policy;

    private List<TreeNode> updateChain;
    private TreeNode rootNode;

    public int treeDepth;
    public int treeSize;
    private double bestQuality;


    public TemporalDifferenceTreeSearchSampler(IProblem problem, int maxLen, Random random, int randomTries, IBanditPolicy policy) {
      this.maxLen = maxLen;
      this.problem = problem;
      this.random = random;
      this.randomTries = randomTries;
      this.policy = policy;
    }

    public void Run(int maxIterations) {
      InitPolicies(problem.Grammar);
      for (int i = 0; !rootNode.done && i < maxIterations; i++) {
        var sentence = SampleSentence(problem.Grammar).ToString();
        var quality = problem.Evaluate(sentence) / problem.BestKnownQuality(maxLen);
        Debug.Assert(quality >= 0 && quality <= 1.0);
        DistributeReward(quality);

        RaiseSolutionEvaluated(sentence, quality);

        if (quality > bestQuality) {
          bestQuality = quality;
          RaiseFoundNewBestSolution(sentence, quality);
        }
      }

      // clean up
      InitPolicies(problem.Grammar); GC.Collect();
    }

    public void PrintStats() {
      var n = rootNode;
      Console.WriteLine("depth: {0,5} size: {1,10} root tries {2,10}, rootQ {3:F3}, bestQ {4:F3}", treeDepth, treeSize, n.tries, n.q, bestQuality);
      while (n.children != null) {
        Console.WriteLine();
        Console.WriteLine("{0,5}->{1,-50}", n.ident, string.Join(" ", n.children.Select(ch => string.Format("{0,4}", ch.ident))));
        Console.WriteLine("{0,5}  {1,-50}", string.Empty, string.Join(" ", n.children.Select(ch => string.Format("{0,4:F2}", ch.q * 10))));
        Console.WriteLine("{0,5}  {1,-50}", string.Empty, string.Join(" ", n.children.Select(ch => string.Format("{0,4}", ch.done ? "X" : ch.tries.ToString()))));
        //n.policy.PrintStats();
        n = n.children.Where(ch => !ch.done).OrderByDescending(c => c.q).First();
      }
      //Console.ReadLine();
    }

    private void InitPolicies(IGrammar grammar) {
      this.updateChain = new List<TreeNode>();

      rootNode = new TreeNode(grammar.SentenceSymbol.ToString());
      treeDepth = 0;
      treeSize = 0;
    }

    private Sequence SampleSentence(IGrammar grammar) {
      updateChain.Clear();
      var startPhrase = new Sequence(grammar.SentenceSymbol);
      return CompleteSentence(grammar, startPhrase);
    }

    private Sequence CompleteSentence(IGrammar g, Sequence phrase) {
      if (phrase.Length > maxLen) throw new ArgumentException();
      if (g.MinPhraseLength(phrase) > maxLen) throw new ArgumentException();
      TreeNode n = rootNode;
      var curDepth = 0;
      while (!phrase.IsTerminal) {
        updateChain.Add(n);

        if (n.randomTries < randomTries) {
          n.randomTries++;
          treeDepth = Math.Max(treeDepth, curDepth);
          return g.CompleteSentenceRandomly(random, phrase, maxLen);
        } else {
          char nt = phrase.FirstNonTerminal;

          int maxLenOfReplacement = maxLen - (phrase.Length - 1); // replacing aAb with maxLen 4 means we can only use alternatives with a minPhraseLen <= 2
          Debug.Assert(maxLenOfReplacement > 0);

          var alts = g.GetAlternatives(nt).Where(alt => g.MinPhraseLength(alt) <= maxLenOfReplacement);

          if (n.randomTries == randomTries && n.children == null) {
            n.children = alts.Select(alt => new TreeNode(alt.ToString())).ToArray(); // create a new node for each alternative
            treeSize += n.children.Length;
          }
          // => select using bandit policy
          int selectedAltIdx = SelectAction(random, n.children);
          Sequence selectedAlt = alts.ElementAt(selectedAltIdx);

          // replace nt with alt
          phrase.ReplaceAt(phrase.FirstNonTerminalIndex, 1, selectedAlt);

          curDepth++;

          // prepare for next iteration
          n = n.children[selectedAltIdx];
        }
      } // while

      updateChain.Add(n);


      // the last node is a leaf node (sentence is done), so we never need to visit this node again
      n.done = true;

      treeDepth = Math.Max(treeDepth, curDepth);
      return phrase;
    }


    // eps-greedy
    private int SelectAction(Random random, TreeNode[] children) {
      if (random.NextDouble() < 0.1) {

        return children.Select((ch, i) => Tuple.Create(ch, i)).Where(p => !p.Item1.done).SelectRandom(random).Item2;
      } else {
        var bestQ = double.NegativeInfinity;
        var bestChildIdx = -1;
        for (int i = 0; i < children.Length; i++) {
          if (children[i].done) continue;
          if (children[i].tries == 0) return i;
          if (children[i].q > bestQ) {
            bestQ = children[i].q;
            bestChildIdx = i;
          }
        }
        Debug.Assert(bestChildIdx > -1);
        return bestChildIdx;
      }
    }

    private void DistributeReward(double reward) {
      const double alpha = 0.1;
      const double gamma = 1;
      // iterate in reverse order (bottom up)
      updateChain.Reverse();
      var nextQ = 0.0;
      foreach (var e in updateChain) {
        var node = e;
        node.tries++;
        if (node.children != null && node.children.All(c => c.done)) {
          node.done = true;
        }
        // reward is recieved only for the last action
        if (e == updateChain.First()) {
          node.q = node.q + alpha * (reward + gamma * nextQ - node.q);
          nextQ = node.q;
        } else {
          node.q = node.q + alpha * (0 + gamma * nextQ - node.q);
          nextQ = node.q;
        }
      }
    }

    private void RaiseSolutionEvaluated(string sentence, double quality) {
      var handler = SolutionEvaluated;
      if (handler != null) handler(sentence, quality);
    }
    private void RaiseFoundNewBestSolution(string sentence, double quality) {
      var handler = FoundNewBestSolution;
      if (handler != null) handler(sentence, quality);
    }
  }
}
